@@ -94,6 +94,7 @@ This is bound dynamically while the checks run.")
           pcase-dolist
           pcase-let
           pcase-let*
+          prog-mode
           read-char-choice
           run-hook-wrapped
           server-eval-at
@@ -136,6 +137,7 @@ This is bound dynamically while the checks run.")
           define-alternative
           define-error
           display-monitor-attributes-list
+          eieio-object-class
           file-acl
           file-extended-attributes
           fill-single-char-nobreak-p
@@ -145,6 +147,7 @@ This is bound dynamically while the checks run.")
           group-real-gid
           hash-table-keys
           hash-table-values
+          line-pixel-height
           macrop
           remove-function
           set-file-acl
@@ -162,6 +165,7 @@ This is bound dynamically while the checks run.")
           window-bottom-divider-width
           window-header-line-height
           window-mode-line-height
+          window-pixel-height
           window-right-divider-width
           window-scroll-bar-width
           window-text-pixel-size
@@ -186,6 +190,8 @@ This is bound dynamically while the checks run.")
           directory-name-p
           file-notify-valid-p
           filepos-to-bufferpos
+          font-lock-ensure
+          font-lock-flush
           format-message
           forward-word-strictly
           frame-edges
@@ -226,7 +232,10 @@ This is bound dynamically while the checks run.")
    (or
     "org-dblock-write:"
     "org-babel-execute:"
-    "org-babel-default-header-args:"))
+    "org-babel-prep-session:"
+    "org-babel-variable-assignments:"
+    "org-babel-default-header-args:"
+    "pcomplete/"))
   "A regexp matching whitelisted non-standard symbol prefixes.")
 
 (defun package-lint--check-all ()
@@ -237,31 +246,88 @@ This is bound dynamically while the checks run.")
       (save-excursion
         (save-restriction
           (widen)
+          (package-lint--check-reserved-keybindings)
           (package-lint--check-keywords-list)
+          (package-lint--check-url-header)
           (package-lint--check-package-version-present)
           (package-lint--check-lexical-binding-is-on-first-line)
+          (package-lint--check-objects-by-regexp
+           "(define-minor-mode\\s-"
+           #'package-lint--check-minor-mode)
+          (package-lint--check-objects-by-regexp
+           "(define-global\\(?:ized\\)?-minor-mode\\s-"
+           #'package-lint--check-globalized-minor-mode)
+          (package-lint--check-objects-by-regexp
+           "(defgroup\\s-" #'package-lint--check-defgroup)
           (let ((desc (package-lint--check-package-el-can-parse)))
             (when desc
               (package-lint--check-package-summary desc)
               (package-lint--check-provide-form desc)))
+          (package-lint--check-no-use-of-cl)
+          (package-lint--check-no-use-of-cl-lib-sublibraries)
+          (package-lint--check-eval-after-load)
           (let ((deps (package-lint--check-dependency-list)))
             (package-lint--check-lexical-binding-requires-emacs-24 deps)
             (package-lint--check-libraries-available-in-emacs deps)
             (package-lint--check-macros-functions-available-in-emacs deps))
           (package-lint--check-for-literal-emacs-path)
+          (package-lint--check-commentary-existence)
           (let ((definitions (package-lint--get-defs)))
             (package-lint--check-autoloads-on-private-functions definitions)
             (package-lint--check-defs-prefix definitions)
             (package-lint--check-symbol-separators definitions)))))
-    package-lint--errors))
+    (sort package-lint--errors
+          (lambda (a b)
+            (pcase-let ((`(,a-line ,a-column ,_ ,a-message) a)
+                        (`(,b-line ,b-column ,_ ,b-message) b))
+              (cond
+               ((/= a-line b-line) (< a-line b-line))
+               ((/= a-column b-column) (< a-column b-column))
+               (t
+                (string-lessp a-message b-message))))))))
 
 (defun package-lint--error (line col type message)
   "Construct a datum for error at LINE and COL with TYPE and MESSAGE."
   (push (list line col type message) package-lint--errors))
 
+(defun package-lint--error-at-point (type message)
+  "Construct a datum for error at the point with TYPE and MESSAGE."
+  (package-lint--error (line-number-at-pos) (current-column) type message))
+
 
 ;;; Checks
 
+(defun package-lint--check-reserved-keybindings ()
+  "Warn about reserved keybindings."
+  (let ((re (rx "(" (*? space) (or "kbd" "global-set-key" "local-set-key" "define-key") symbol-end)))
+    (goto-char (point-min))
+    (while (re-search-forward re nil t)
+      (unless (nth 8 (save-match-data (syntax-ppss)))
+        ;; Read form and get key-sequence
+        (goto-char (match-beginning 0))
+        (let ((seq (package-lint--extract-key-sequence
+                    (read (current-buffer)))))
+          (when seq
+            (let ((message (package-lint--test-keyseq seq)))
+              (when message
+                (package-lint--error-at-point 'warning message)))))))))
+
+(defun package-lint--check-commentary-existence ()
+  "Warn about nonexistent or empty commentary section."
+  (let ((start (lm-commentary-start)))
+    (if (null start)
+        (package-lint--error
+         1 1 'error
+         "Package should have a ;;; Commentary section.")
+      ;; Skip over the section header.
+      (goto-char start)
+      (forward-line)
+      (when (package-lint--region-empty-p (point) (lm-commentary-end))
+        (goto-char start)
+        (package-lint--error-at-point
+         'error
+         "Package should have a non-empty ;;; Commentary section.")))))
+
 (defun package-lint--check-autoloads-on-private-functions (definitions)
   "Verify that private functions DEFINITIONS don't have autoload cookies."
   (pcase-dolist (`(,symbol . ,position) definitions)
@@ -269,19 +335,22 @@ This is bound dynamically while the checks run.")
       (goto-char position)
       (forward-line -1)
       (when (looking-at-p (rx ";;;###autoload"))
-        (package-lint--error
-         (line-number-at-pos) (current-column) 'warning
+        (package-lint--error-at-point
+         'warning
          "Private functions generally should not be autoloaded.")))))
 
 (defun package-lint--check-for-literal-emacs-path ()
   "Verify package does not refer to \"\.emacs\.d\" literally.
 Instead it should use `user-emacs-directory' or `locate-user-emacs-file'."
   (goto-char (point-min))
-  (while (re-search-forward "\\.emacs\\.d" nil t)
+  ;; \b won't find a word boundary between a symbol and the "." in
+  ;; ".emacs.d". / is a valid symbol constituent in Emacs Lisp, so
+  ;; must be explicitly blacklisted.
+  (while (re-search-forward "\\(?:\\_<\\|/\\)\\.emacs\\.d\\b" nil t)
     (unless (nth 4 (syntax-ppss))
       ;; Not in a comment
-      (package-lint--error
-       (line-number-at-pos) (current-column) 'warning
+      (package-lint--error-at-point
+       'warning
        "Use variable `user-emacs-directory' or function `locate-user-emacs-file' instead of a literal path to the Emacs user directory or files."))))
 
 (defun package-lint--check-keywords-list ()
@@ -294,6 +363,22 @@ Instead it should use `user-emacs-directory' or `locate-user-emacs-file'."
          line-no 1 'warning
          (format "You should include standard keywords: see the variable `finder-known-keywords'."))))))
 
+(defun package-lint--check-url-header ()
+  "Verify that the package has an HTTPS or HTTP Homepage/URL header."
+  (if (package-lint--goto-header "\\(?:URL\\|Homepage\\)")
+      (let ((url (match-string 3))
+            (url-start (match-beginning 3)))
+        (unless (and (equal (thing-at-point 'url) url)
+                     (string-match-p "^https?://" url))
+          (package-lint--error
+           (line-number-at-pos)
+           (1+ (- url-start (line-beginning-position)))
+           'error
+           "Package URLs should be a single HTTPS or HTTP URL.")))
+    (package-lint--error
+     1 1 'error
+     "Package should have a Homepage or URL header.")))
+
 (defun package-lint--check-dependency-list ()
   "Check the contents of the \"Package-Requires\" header.
 Return a list of well-formed dependencies, same as
@@ -359,6 +444,20 @@ the form (PACKAGE-NAME PACKAGE-VERSION LINE-NO LINE-BEGINNING-OFFSET)."
           (format "Expected (package-name \"version-num\"), but found %S." entry)))))
     valid-deps))
 
+(defun package-lint--check-package-installable (archive-entry package-version line-no offset)
+  "Check that ARCHIVE-ENTRY is installable from a configured package archive.
+
+Check that package described by ARCHIVE-ENTRY can be installed at
+required version PACKAGE-VERSION.  If not, raise an error for
+LINE-NO at OFFSET."
+  (let* ((package-name (car archive-entry))
+         (best-version (package-lint--highest-installable-version-of package-name)))
+    (when (version-list-< best-version package-version)
+      (package-lint--error
+       line-no offset 'warning
+       (format "Version dependency for %s appears too high: try %s" package-name
+               (package-version-join best-version))))))
+
 (defun package-lint--check-packages-installable (valid-deps)
   "Check that all VALID-DEPS are available for installation."
   (pcase-dolist (`(,package-name ,package-version ,line-no ,offset) valid-deps)
@@ -370,12 +469,7 @@ the form (PACKAGE-NAME PACKAGE-VERSION LINE-NO LINE-BEGINNING-OFFSET)."
       ;; Not 'emacs
       (let ((archive-entry (assq package-name package-archive-contents)))
         (if archive-entry
-            (let ((best-version (package-lint--lowest-installable-version-of package-name)))
-              (when (version-list-< best-version package-version)
-                (package-lint--error
-                 line-no offset 'warning
-                 (format "Version dependency for %s appears too high: try %s" package-name
-                         (package-version-join best-version)))))
+            (package-lint--check-package-installable archive-entry package-version line-no offset)
           (package-lint--error
            line-no offset 'error
            (format "Package %S is not installable." package-name)))))))
@@ -409,8 +503,21 @@ the form (PACKAGE-NAME PACKAGE-VERSION LINE-NO LINE-BEGINNING-OFFSET)."
          lexbind-line lexbind-col 'warning
          "You should depend on (emacs \"24\") if you need lexical-binding.")))))
 
+(defun package-lint--inside-comment-or-string-p ()
+  "Return non-nil if point is inside a comment or string."
+  (let ((ppss (save-match-data (syntax-ppss))))
+    (or (nth 3 ppss) (nth 4 ppss))))
+
+(defun package-lint--seen-fboundp-check-for (sym)
+  "Return non-nil if a `fboundp' check for SYM is present before point."
+  (save-excursion
+    (save-match-data
+      (and (re-search-backward
+            (concat "(fboundp\\s-+'" (regexp-quote sym) "\\_>") (point-min) t)
+           (not (package-lint--inside-comment-or-string-p))))))
+
 (defun package-lint--check-version-regexp-list (valid-deps list rx-start rx-end)
-  "Warn about matches of REGEXP when VERSION is not in VALID-DEPS.
+  "Warn about any match of REGEXP when VERSION is not in VALID-DEPS.
 LIST is an alist of (VERSION . REGEXP*).
 REGEXP is (concat RX-START REGEXP* RX-END) for each REGEXP*."
   (let ((emacs-version-dep (or (cadr (assq 'emacs valid-deps)) '(0))))
@@ -418,15 +525,46 @@ REGEXP is (concat RX-START REGEXP* RX-END) for each REGEXP*."
       (when (version-list-< emacs-version-dep added-in-version)
         (goto-char (point-min))
         (while (re-search-forward (concat rx-start regexp rx-end) nil t)
-          (unless (let ((ppss (save-match-data (syntax-ppss))))
-                    (or (nth 3 ppss) (nth 4 ppss)))
-            (package-lint--error
-             (line-number-at-pos)
-             (save-excursion (goto-char (match-beginning 1)) (current-column))
-             'error
-             (format "You should depend on (emacs \"%s\") if you need `%s'."
-                     (mapconcat #'number-to-string added-in-version ".")
-                     (match-string-no-properties 1)))))))))
+          (unless (package-lint--inside-comment-or-string-p)
+            (let ((sym (match-string-no-properties 1)))
+              (unless (package-lint--seen-fboundp-check-for sym)
+                (save-excursion
+                  (goto-char (match-beginning 1))
+                  (package-lint--error-at-point
+                   'error
+                   (format "You should depend on (emacs \"%s\") if you need `%s'."
+                           (mapconcat #'number-to-string added-in-version ".")
+                           sym)))))))))))
+
+(defun package-lint--check-eval-after-load ()
+  "Warn about use of `eval-after-load' and co."
+  (save-excursion
+    (save-match-data
+      (goto-char (point-min))
+      (when (re-search-forward "(\\s-*?\\(\\(?:with-\\)?eval-after-load\\)\\_>" nil t)
+        (package-lint--error-at-point
+         'warning
+         (format "`%s' is for use in configurations, and should rarely be used in packages." (match-string 1)))))))
+
+(defun package-lint--check-no-use-of-cl ()
+  "Warn about use of deprecated `cl' library."
+  (save-excursion
+    (save-match-data
+      (goto-char (point-min))
+      (when (re-search-forward "(\\s-*?require\\s-*?'cl\\_>" nil t)
+        (package-lint--error-at-point
+         'warning
+         "Replace deprecated `cl' with `cl-lib'.  The `cl-libify' package can help with this.")))))
+
+(defun package-lint--check-no-use-of-cl-lib-sublibraries ()
+  "Warn about use of `cl-macs', `cl-seq' etc."
+  (save-excursion
+    (save-match-data
+      (goto-char (point-min))
+      (when (re-search-forward "(\\s-*?require\\s-*?'cl-\\(?:macs\\|seq\\)\\_>" nil t)
+        (package-lint--error-at-point
+         'warning
+         "This file is not in the `cl-lib' ELPA compatibility package: require `cl-lib' instead.")))))
 
 (defun package-lint--check-libraries-available-in-emacs (valid-deps)
   "Warn about use of libraries that are not available in the Emacs version in VALID-DEPS."
@@ -550,7 +688,7 @@ DESC is a struct as returned by `package-buffer-info'."
        "The package summary is too long. It should be at most 50 characters.")))
     (when (save-match-data
             (let ((case-fold-search t))
-              (and (string-match "\\<emacs\\>" summary)
+              (and (string-match "[^.]\\<emacs\\>" summary)
                    (not (string-match-p "[[:space:]]+lisp" summary (match-end 0))))))
       (package-lint--error
        1 1
@@ -583,29 +721,139 @@ DESC is a struct as returned by `package-buffer-info'."
            (format "`%s' contains a non-standard separator `%s', use hyphens instead (see Elisp Coding Conventions)."
                    name (substring-no-properties name match-pos (1+ match-pos)))))))))
 
+(defun package-lint--valid-definition-name-p (name prefix-re position)
+  "Return non-nil if NAME denotes a valid definition name.
+
+Valid definition names are:
+
+- a NAME starting with PREFIX-RE, a regular expression
+  representing the current package prefix,
+
+- a NAME matching `package-lint--sane-prefixes', or
+
+- a NAME whose POSITION in the buffer denotes a global definition."
+  (or (string-match-p prefix-re name)
+      (string-match-p package-lint--sane-prefixes name)
+      (progn
+        (goto-char position)
+        (looking-at-p (rx (*? space) "(" (*? space)
+                          (or "defadvice" "cl-defmethod")
+                          symbol-end)))))
+
 (defun package-lint--check-defs-prefix (definitions)
   "Verify that symbol DEFINITIONS start with package prefix."
   (let ((prefix (package-lint--get-package-prefix)))
     (when prefix
-      (let ((prefix-re (rx-to-string `(seq string-start ,prefix (or "-" string-end)))))
+      (let ((prefix-re
+             (rx-to-string
+              `(seq string-start
+                    (or (seq ,prefix (or "-" string-end))
+                        (seq "global-"
+                             ,prefix
+                             (or "-mode"
+                                 (seq "-" (* any) "-mode"))
+                             string-end))))))
         (pcase-dolist (`(,name . ,position) definitions)
-          (unless (or (string-match-p prefix-re name)
-                      (string-match-p package-lint--sane-prefixes name))
+          (unless (package-lint--valid-definition-name-p name prefix-re position)
             (let ((line-no (line-number-at-pos position)))
               (package-lint--error
                line-no 1 'error
                (format "\"%s\" doesn't start with package's prefix \"%s\"."
                        name prefix)))))))))
 
+(defun package-lint--check-minor-mode (def)
+  "Offer up concerns about the minor mode definition DEF."
+  (when (cl-search '(:global t) def)
+    (package-lint--check-globalized-minor-mode def)))
+
+(defun package-lint--check-globalized-minor-mode (def)
+  "Offer up concerns about the global minor mode definition DEF."
+  (let ((feature (intern (package-lint--provided-feature)))
+        (autoloaded (save-excursion
+                      (forward-line -1)
+                      (beginning-of-line)
+                      (looking-at ";;;###autoload"))))
+    (unless (or autoloaded
+                (cl-search `(:require ',feature) def :test #'equal))
+      (package-lint--error-at-point
+       'error
+       (format
+        "Global minor modes should be autoloaded or, rarely, `:require' their defining file (i.e. \":require '%s\"), to support the customization variable of the same name." feature)))))
+
+(defun package-lint--check-defgroup (def)
+  "Offer up concerns about the customization group definition DEF."
+  (when (symbolp (cadr def))
+    (let ((group-name (symbol-name (cadr def))))
+      (when (string-match "\\(.*\\)-mode$" group-name)
+        (let ((parent (intern (match-string 1 group-name))))
+          (unless (cl-search `(:group ',parent) def :test #'equal)
+            (package-lint--error-at-point
+             'error
+             "Customization groups should not end in \"-mode\" unless that name would conflict with their parent group."))))))
+
+  (unless (memq :group def)
+    (package-lint--error-at-point
+     'error
+     "Customization groups should specify a parent via `:group'.")))
+
 
 ;;; Helpers
 
-(defun package-lint--lowest-installable-version-of (package)
-  "Return the lowest version of PACKAGE available for installation."
+(defun package-lint--extract-key-sequence (form)
+  "Extract the key sequence from FORM."
+  (pcase form
+    (`(kbd ,seq)
+     (package-lint--extract-key-sequence seq))
+    ((or `(global-set-key ,seq ,_) `(local-set-key ,seq ,_))
+     (package-lint--extract-key-sequence seq))
+    (`(define-key ,_ ,seq ,_)
+     (package-lint--extract-key-sequence seq))
+    ((pred stringp)
+     (listify-key-sequence (read-kbd-macro form)))
+    ((pred vectorp)
+     (unless (listp (elt form 0))
+       (listify-key-sequence form)))))
+
+(defun package-lint--test-keyseq (lks)
+  "Return a message if the listified key sequence LKS is invalid, otherwise nil."
+  (let* ((modifiers (event-modifiers lks))
+         (basic-type (event-basic-type lks)))
+    (when (or (and (> (length lks) 1) (equal (car (last lks)) ?\C-g))
+              (and (equal (car (last lks)) ?\e)
+                   (not (equal (nthcdr (- (length lks) 2) lks)
+                               '(?\e ?\e))))
+              (equal (car (last lks)) ?\C-h)
+              (and (equal modifiers '(control))
+                   (equal ?c basic-type)
+                   (cdr lks)
+                   (let ((v (event-basic-type (cdr lks)))
+                         (m (event-modifiers (cdr lks))))
+                     (and (numberp v)
+                          (<= v ?z)
+                          (>= v ?a)
+                          (or (null m) (equal '(shift) m)))))
+              (member basic-type '(f5 f6 f7 f8 f9)))
+      "This key sequence is reserved (see Key Binding Conventions in the Emacs Lisp manual)")))
+
+(defun package-lint--region-empty-p (start end)
+  "Return t iff the region between START and END has no non-empty lines.
+
+Lines consisting only of whitespace or empty comments are considered empty."
+  (save-excursion
+    (save-restriction
+      (let ((inhibit-changing-match-data t))
+        (narrow-to-region start end)
+        (goto-char start)
+        (while (and (looking-at "^[[:space:]]*;*[[:space:]]*$")
+                    (= 0 (forward-line))))
+        (eobp)))))
+
+(defun package-lint--highest-installable-version-of (package)
+  "Return the highest version of PACKAGE available for installation."
   (let ((descriptors (cdr (assq package package-archive-contents))))
     (if (fboundp 'package-desc-version)
         (car (sort (mapcar 'package-desc-version descriptors)
-                   #'version-list-<))
+                   (lambda (v1 v2) (not (version-list-< v1 v2)))))
       (aref descriptors 0))))
 
 (defun package-lint--goto-header (header-name)
@@ -693,6 +941,23 @@ Prefix is returned without any `-mode' suffix."
     (when feature
       (replace-regexp-in-string "-mode\\'" "" feature))))
 
+(defun package-lint--check-objects-by-regexp (regexp function)
+  "Check all objects with the literal printed form matching REGEXP.
+
+The objects are parsed with `read'.  The FUNCTION is passed the
+read object, with the point at the beginning of the match.
+
+S-expressions in comments or comments, partial s-expressions, or
+otherwise invalid read forms are ignored."
+  (goto-char (point-min))
+  (while (re-search-forward regexp nil t)
+    (save-excursion
+      (goto-char (match-beginning 0))
+      (let ((obj (unless (package-lint--inside-comment-or-string-p)
+                   (save-excursion
+                     (ignore-errors (read (current-buffer)))))))
+        (when obj (funcall function obj))))))
+
 
 ;;; Public interface
 
@@ -716,18 +981,20 @@ Current buffer is used if none is specified."
   (interactive)
   (let ((errs (package-lint-buffer))
         (buf "*Package-Lint*"))
-    (with-current-buffer (get-buffer-create buf)
-      (let ((buffer-read-only nil))
-        (erase-buffer)
-        (cond
-         ((null errs) (insert "No issues found."))
-         ((null (cdr errs)) (insert "1 issue found:\n\n"))
-         (t (insert (format "%d issues found:\n\n" (length errs)))))
-        (pcase-dolist (`(,line ,col ,type ,message) errs)
-          (insert (format "%d:%d: %s: %s\n" line col type message))))
-      (special-mode)
-      (view-mode 1))
-    (display-buffer buf)))
+    (if (null errs)
+        (message "No issues found")
+      (with-current-buffer (get-buffer-create buf)
+        (let ((buffer-read-only nil))
+          (erase-buffer)
+          (cond
+           ((null errs) (insert "No issues found."))
+           ((null (cdr errs)) (insert "1 issue found:\n\n"))
+           (t (insert (format "%d issues found:\n\n" (length errs)))))
+          (pcase-dolist (`(,line ,col ,type ,message) errs)
+            (insert (format "%d:%d: %s: %s\n" line col type message))))
+        (special-mode)
+        (view-mode 1))
+      (display-buffer buf))))
 
 ;;;###autoload
 (defun package-lint-batch-and-exit ()
@@ -767,4 +1034,7 @@ otherwise."
          nil t)))))
 
 (provide 'package-lint)
+;; Local Variables:
+;; indent-tabs-mode: nil
+;; End:
 ;;; package-lint.el ends here
